@@ -1,10 +1,17 @@
-import boto3
+# SPDX-FileCopyrightText: Mintlab B.V.
+#
+# SPDX-License-Identifier: EUPL-1.2
+
+import minio
+import minio.credentials
 import minty.infrastructure.mime_utils as mu
 import os
-from botocore.client import Config
+import urllib.parse
+from datetime import timedelta
 from minty import Base
 from minty.exceptions import ConfigurationConflict
 from threading import BoundedSemaphore
+from typing import BinaryIO, Optional
 from uuid import UUID
 
 max_concurrent_parsers = 1
@@ -28,62 +35,64 @@ class S3Wrapper(Base):
             f"No configuration found for '{storage_location}'"
         )
 
-    def _generate_attachment_disposition(self, filename: str, download: bool):
+    def _generate_attachment_disposition(
+        self, filename: Optional[str], download: bool
+    ):
         """Generate the value of the Content-Disposition header
 
         https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition
         """
         if download:
-            content_disposition = f'attachment; filename="{filename}"'
+            content_disposition = "attachment"
         else:
-            content_disposition = f'inline; filename="{filename}"'
+            content_disposition = "inline"
+
+        if filename is not None:
+            encoded_filename = urllib.parse.quote(filename)
+            content_disposition += f"; filename*=UTF-8''{encoded_filename}"
+
         return content_disposition
 
-    def _s3_client(self, config):
+    def _s3_client(self, config) -> minio.Minio:
         """Build an AWS S3 client object from configuration"""
 
-        boto_config = self._generate_boto_config(config)
-        boto_parameters = self._generate_boto_parameters(config)
+        minio_config = self._generate_minio_config(config)
+        return minio.Minio(**minio_config)
 
-        return boto3.session.Session().client(
-            service_name="s3", config=Config(**boto_config), **boto_parameters
-        )
+    def _generate_minio_config(self, config):
+        """Generate arguments to set up a Minio instance, using the configuration."""
 
-    def _generate_boto_parameters(self, config):
-        """Generate the boto parameters from the given config."""
+        minio_parameters = {}
 
-        boto_parameters = {}
         access_key = config.get("access_id", None)
         secret_key = config.get("access_key", None)
         endpoint_url = config.get("endpoint_url", None)
         region_name = config.get("region_name", None)
 
-        # If access id and key are not specified, boto3 will default to use
-        # the EC2 IAM role, which is what we want on production.
+        # If access id and key are not specified, Minio needs to be set up to use
+        # will default to use the EC2 IAM role (which is what we want on
+        # production)
 
         if all([access_key, secret_key]):
-            boto_parameters["aws_access_key_id"] = access_key
-            boto_parameters["aws_secret_access_key"] = secret_key
+            minio_parameters["access_key"] = access_key
+            minio_parameters["secret_key"] = secret_key
+        else:
+            minio_parameters[
+                "credentials"
+            ] = minio.credentials.IamAwsProvider()
 
         if endpoint_url is not None:
-            boto_parameters["endpoint_url"] = endpoint_url
+            if endpoint_url.startswith("https://"):
+                minio_parameters["secure"] = True
+                minio_parameters["endpoint"] = endpoint_url[len("https://") :]
+            else:
+                minio_parameters["secure"] = False
+                minio_parameters["endpoint"] = endpoint_url[len("http://") :]
 
         if region_name is not None:
-            boto_parameters["region_name"] = region_name
+            minio_parameters["region"] = region_name
 
-        return boto_parameters
-
-    def _generate_boto_config(self, config):
-        """Generate the boto config from the given config."""
-
-        signature_version = config.get("signature_version", None)
-        addressing_style = config.get("addressing_style", "auto")
-
-        boto_config = {"s3": {"addressing_style": addressing_style}}
-        if signature_version is not None:
-            boto_config["signature_version"] = signature_version
-
-        return boto_config
+        return minio_parameters
 
     def _get_s3_bucket_and_key(self, config: dict, uuid: UUID) -> dict:
         """Generate `key` and `bucket` values for s3 based on the config type.
@@ -112,39 +121,47 @@ class S3Wrapper(Base):
             )
         return {"key": key, "bucket": bucket}
 
-    def upload(self, file_handle, uuid):
+    def upload(
+        self, file_handle: BinaryIO, uuid, file_size: Optional[int] = None
+    ):
+        """Uploads data from a file handle to S3"""
 
-        timer = self.statsd.get_timer("file_upload")
-        with timer.time("s3_upload"):
+        if file_size is None:
+            file_handle.seek(0, os.SEEK_END)
+            file_size = file_handle.tell()
+            file_handle.seek(0, os.SEEK_SET)
+
+        timer = self.statsd.get_timer("filestorage_write_duration")
+        with timer.time():
             config = self._select_config()
 
             location_name = config["name"]
             s3_config = self._get_s3_bucket_and_key(config=config, uuid=uuid)
 
-            response_dict = self._s3_client(config=config).put_object(
-                Body=file_handle,
-                Bucket=s3_config["bucket"],
-                Key=s3_config["key"],
+            put_result = self._s3_client(config=config).put_object(
+                bucket_name=s3_config["bucket"],
+                object_name=s3_config["key"],
+                data=file_handle,
+                length=file_size,
             )
-
-        file_handle.seek(0, os.SEEK_END)
-        total_size = file_handle.tell()
+            etag = put_result.etag
+        self.statsd.get_counter("filestorage_write_number").increment()
 
         global mime_parser_semaphore
         with mime_parser_semaphore:
             mimetype = mu.get_mime_type_from_handle(file_handle)
 
-        var = {
+        return {
             "uuid": uuid,
-            "md5": response_dict["ETag"][1:-1],
-            "size": total_size,
+            "md5": etag[1:-1],
+            "size": file_size,
             "mime_type": mimetype,
             "storage_location": location_name,
         }
 
-        return var
-
-    def download_file(self, destination, file_uuid, storage_location: str):
+    def download_file(
+        self, destination: BinaryIO, file_uuid, storage_location: str
+    ):
         """Download a file from S3 to the file-handle `destination`
 
         :param destination: File handle to write the file's contents to
@@ -155,16 +172,27 @@ class S3Wrapper(Base):
             this file.
         :type storage_location: str
         """
-        timer = self.statsd.get_timer("file_download")
+
         config = self._select_config(storage_location)
         s3_config = self._get_s3_bucket_and_key(config=config, uuid=file_uuid)
 
-        with timer.time("s3_download"):
-            self._s3_client(config=config).download_fileobj(
-                s3_config["bucket"], s3_config["key"], destination
-            )
+        timer = self.statsd.get_timer("filestorage_read_duration")
+        with timer.time():
+            file_data = None
+            try:
+                file_data = self._s3_client(config=config).get_object(
+                    bucket_name=s3_config["bucket"],
+                    object_name=s3_config["key"],
+                )
+                for chunk in file_data.stream(128 * 1024):
+                    destination.write(chunk)
+            finally:
+                if file_data:
+                    file_data.close()
+                    file_data.release_conn()
 
         destination.flush()
+        self.statsd.get_counter("filestorage_read_number").increment()
 
         return
 
@@ -172,35 +200,40 @@ class S3Wrapper(Base):
         self,
         uuid,
         storage_location: str,
-        filename: str,
+        filename: Optional[str],
         mime_type: str,
         download: bool,
     ):
         """Generate a pre-signed download url for given file uuid."""
-        timer = self.statsd.get_timer("file_download_url")
         config = self._select_config(storage_location)
         s3_config = self._get_s3_bucket_and_key(config=config, uuid=uuid)
 
         presigned_url_expiration = config.get("presigned_url_expiration", 3600)
         cache_control = config.get("cache_control", "private, max-age=3600")
 
-        with timer.time("s3_download_url"):
-            content_disposition = self._generate_attachment_disposition(
-                filename=filename, download=download
-            )
+        content_disposition = self._generate_attachment_disposition(
+            filename=filename, download=download
+        )
+
+        timer = self.statsd.get_timer(
+            "filestore_generate_tmp_read_link_duration"
+        )
+        with timer.time():
             presigned_url = self._s3_client(
                 config=config
-            ).generate_presigned_url(
-                "get_object",
-                Params={
-                    "Bucket": s3_config["bucket"],
-                    "Key": s3_config["key"],
-                    "ResponseContentDisposition": content_disposition,
-                    "ResponseContentType": mime_type,
-                    "ResponseCacheControl": cache_control,
+            ).presigned_get_object(
+                bucket_name=s3_config["bucket"],
+                object_name=s3_config["key"],
+                expires=timedelta(seconds=presigned_url_expiration),
+                response_headers={
+                    "response-content-disposition": content_disposition,
+                    "response-content-type": mime_type,
+                    "response-cache-control": cache_control,
                 },
-                ExpiresIn=presigned_url_expiration,
             )
+        self.statsd.get_counter(
+            "filestore_generate_tmp_read_link_number"
+        ).increment()
 
         return presigned_url
 
